import numpy as np

from multiprocessing import Pool, cpu_count
from tqdm.auto import tqdm, trange


def make_sample_n(sample_fn, parallel=True, pool=None):
    """Takes in a sample function and allows for parallelized sampling

    :param sample_fn: A function to compute samples
    :type sample_fn: function
    :param parallel: True if parallelization is to be used to compute samples, defaults to True
    :type parallel: bool, optional
    :param pool: Pool to use for parallelization if specified, defaults to None
    :type pool: multiprocessing.pool.Pool, optional
    """

    def sample_n(n, pool=pool):
        """Inner function to draw n samples using a specified sample function

        :param n: The number of samples to compute
        :type n: int
        :param pool: Pool to use for parallelization if specified, defaults to pool
        :type pool: multiprocessing.pool.Pool, optional
        :return: Array of n samples
        :rtype: numpy.ndarray
        """
        if parallel:
            if pool is None:
                print(f"Using {cpu_count()} CPUs")
                p = Pool(cpu_count())
            else:
                p = pool
            return np.array([s for s in tqdm(p.imap(sample_fn, np.arange(n)), total=n)])
        else:
            return np.array([sample_fn() for i in trange(n)])

    return sample_n



import numpy as np

from collections.abc import Callable, Iterable
from functools import partial
from numpy.random import default_rng
from sklearn.preprocessing import PolynomialFeatures


class ScalarDisturbance:
    """Class that models the scalar disturbance in a single dimension of a dynamic system

    :param basis_funcs: The functions factored into the scalar disturbance
    :type basis_funcs: list
    :param num_funcs: The number of basis functions
    :type num_funcs: int
    :param map_funcs: If true, `basis_funcs` is assumed to be a single function that plays the role of multiple basis functions when an input is passed to it, defaults to False
    :type map_funcs: bool, optional
    :param decay_weights: If True, the weights of the basis functions decrease as the index of the functions increase, defaults to False
    :type decay_weights: bool, optional
    :param negative_weights: If False, the weights of the functions must only be positive, defaults to True
    :type negative_weights: bool, optional
    """

    def __init__(
        self,
        basis_funcs,
        num_funcs,
        map_funcs=False,
        decay_weights=False,
        negative_weights=True,
    ):
        self.basis_funcs = basis_funcs
        self.num_funcs = num_funcs
        self.map_funcs = map_funcs
        self.decay_weights = decay_weights
        self.negative_weights = negative_weights

        self.check_funcs()
        self.ru = default_rng().uniform

        if not self.decay_weights:
            self.upper_alpha = 1
            if self.negative_weights:
                self.lower_alpha = -1
            else:
                self.lower_alpha = 0
        else:
            arr_upper = np.arange(1, self.num_funcs + 1)
            self.upper_alpha = 1 / arr_upper
            if self.negative_weights:
                self.lower_alpha = -self.upper_alpha
            else:
                self.lower_alpha = np.zeros(self.num_funcs)

        # by default functions are equally weighted
        self.alpha = np.ones(self.num_funcs)

    @classmethod
    def from_iterable(cls, basis_funcs, decay_weights=False, negative_weights=True):
        """Class method that allows for an instance of :class:`dadra.ScalarDisturbance` to be initialized directly from a list of functions

        :param basis_funcs: A list of basis functions to be factored into the scalar disturbance
        :type basis_funcs: list
        :param decay_weights: If True, the weights of the basis functions decrease as the index of the functions increase, defaults to False
        :type decay_weights: bool, optional
        :param negative_weights: If False, the weights of the functions must only be positive, defaults to True
        :type negative_weights: bool, optional
        :return: A :class:`dadra.ScalarDisturbance` object
        :rtype: :class:`dadra.ScalarDisturbance`
        """
        num_funcs = len(basis_funcs)
        return cls(
            basis_funcs,
            num_funcs,
            map_funcs=False,
            decay_weights=decay_weights,
            negative_weights=negative_weights,
        )

    @classmethod
    def sin_disturbance(cls, num_funcs, decay_weights=False, negative_weights=True):
        """Class method that allows for an instance of :class:`dadra.ScalarDisturbance` to be initialized directly from a list of sinusoidal basis functions

        :param num_funcs: The number of basis functions to include in the scalar disturbance
        :type num_funcs: int
        :param decay_weights: If True, the weights of the basis functions decrease as the index of the functions increase, defaults to False
        :type decay_weights: bool, optional
        :param negative_weights: If False, the weights of the functions must only be positive, defaults to True
        :type negative_weights: bool, optional
        :return: A :class:`dadra.ScalarDisturbance` object
        :rtype: :class:`dadra.ScalarDisturbance`
        """
        basis_funcs = [cls.sinusoid(k) for k in range(num_funcs)]
        return cls(
            basis_funcs,
            num_funcs,
            map_funcs=False,
            decay_weights=decay_weights,
            negative_weights=negative_weights,
        )

    @classmethod
    def poly_disturbance(cls, num_funcs, decay_weights=True, negative_weights=True):
        """Class method that allows for an instance of :class:`dadra.ScalarDisturbance` to be initialized directly from a list of polynomial basis functions

        :param num_funcs: The number of basis functions to include in the scalar disturbance
        :type num_funcs: int
        :param decay_weights: If True, the weights of the basis functions decrease as the index of the functions increase, defaults to False
        :type decay_weights: bool, optional
        :param negative_weights: If False, the weights of the functions must only be positive, defaults to True
        :type negative_weights: bool, optional
        :return: A :class:`dadra.ScalarDisturbance` object
        :rtype: :class:`dadra.ScalarDisturbance`
        """
        deg = num_funcs - 1
        basis_funcs = cls.poly(deg)
        return cls(
            basis_funcs,
            num_funcs,
            map_funcs=True,
            decay_weights=decay_weights,
            negative_weights=negative_weights,
        )

    @classmethod
    def no_disturbance(cls):
        """Class method that allows for an instance of :class:`dadra.ScalarDisturbance` to be initialized with no disturbance

        :return: A :class:`dadra.ScalarDisturbance` object
        :rtype: :class:`dadra.ScalarDisturbance`
        """
        basis_func = [cls.zero_func]
        return cls(basis_func, 1)

    @staticmethod
    def sinusoid(k):
        """Defines the k-th sinusoidal basis function of a scalar disturbance

        :param k: The index of the basis function
        :type k: int
        :return: The k-th sinusoidal basis function
        :rtype: function
        """
        if k == 0:
            return ScalarDisturbance.k_sin_one
        else:
            k_sin = partial(ScalarDisturbance.sin_basis, k=k)
            return k_sin

    @staticmethod
    def k_sin_one(t):
        """Constant function that returns 1

        :param t: The time
        :type t: float
        :return: 1
        :rtype: int
        """
        return 1

    @staticmethod
    def sin_basis(t, k):
        """The k-th sinusoidal basis function

        :param t: The time
        :type t: float
        :return: The output of the basis function at time t
        :rtype: float
        """
        return np.sin(2 * np.pi * k * t)

    @staticmethod
    def poly(deg):
        """Defines the mapping function which computes all basis functions for a polynomial disturbance

        :param deg: The maximum degree polynomial
        :type deg: int
        :return: The poylynomial mapping function
        :rtype: function
        """
        poly_map = PolynomialFeatures(degree=deg).fit_transform
        return poly_map

    @staticmethod
    def zero_func(t):
        """Defines the basis function for zero disturbance

        :param t: The time
        :type t: float
        :return: [description]
        :rtype: [type]
        """
        return 0

    def check_funcs(self):
        """Checks that the basis functions are all valid

        :raises TypeError: If 'basis_funcs' is not an Iterable when 'map_funcs' is set to False
        :raises TypeError: If the items within 'basis_funcs' are not callable functions when 'map_funcs' is set to False
        :raises TypeError: If 'basis_funcs' is not a callable function when 'map_funcs' is set to True
        """
        if not self.map_funcs:
            if not isinstance(self.basis_funcs, Iterable):
                raise TypeError(
                    "'ScalarDisturbance' object attribute 'basis_funcs' must be an Iterable if attribute 'map_funcs' is False"
                )
            for func in self.basis_funcs:
                if not isinstance(func, Callable):
                    raise TypeError(
                        "'ScalarDisturbance' object attribute 'basis_funcs' must consist of callable functions if attribute 'map_funcs' is False"
                    )
        else:
            if not isinstance(self.basis_funcs, Callable):
                raise TypeError(
                    "'ScalarDisturbance' object attribute 'basis_funcs' must be a callable function if attribute 'map_funcs' is True"
                )

    def draw_alpha(self):
        """Initializes alpha, the weights for the basis functions, from an m-dimensional random distribution"""
        self.alpha = self.ru(self.lower_alpha, self.upper_alpha, size=(self.num_funcs,))

    def d(self, t):
        """Computes the scalar disturbance at the specified time

        :param t: The time
        :type t: float
        :return: The scalar disturbance
        :rtype: float
        """
        if not self.map_funcs:
            weighted_funcs = [a * f(t) for a, f in zip(self.alpha, self.basis_funcs)]
            return sum(weighted_funcs)
        else:
            t_arr = np.array([[t]])
            y = self.basis_funcs(t_arr)
            return float(np.dot(y, self.alpha))


class Disturbance:
    """Class that models the disturbance for all n-dimensions of a dynamic system

    :param dist_list: A list of objects of type :class:`dadra.ScalarDisturbance`
    :type dist_list: list
    """

    def __init__(self, dist_list):
        self.dist_list = dist_list
        self.num_dists = len(dist_list)

    @classmethod
    def n_sin_disturbance(
        cls, n, num_funcs=5, decay_weights=False, negative_weights=True
    ):
        """Class method that creates an instance of :class:`dadra.Disturbance` consisting of sinusoidal disturbance for each dimension

        :param n: The dimension of the disturbance, corresponding to the state dimension of the dynamic system
        :type n: int
        :param num_funcs: The number of basis functions (m) for each scalar dimension, defaults to 5
        :type num_funcs: int, optional
        :param decay_weights: If True, the weights of the basis functions decrease as the index of the functions increase, defaults to False
        :type decay_weights: bool, optional
        :param negative_weights: If False, the weights of the basis functions must only be positive, defaults to True
        :type negative_weights: bool, optional
        :return: A :class:`dadra.Disturbance` object
        :rtype: :class:`dadra.Disturbance`
        """
        dist_list = [
            ScalarDisturbance.sin_disturbance(
                num_funcs,
                decay_weights=decay_weights,
                negative_weights=negative_weights,
            )
            for _ in range(n)
        ]
        return cls(dist_list)

    @classmethod
    def n_poly_disturbance(
        cls, n, num_funcs=5, decay_weights=True, negative_weights=True
    ):
        """Class method that creates an instance of :class:`dadra.Disturbance` consisting of polynomial disturbance for each dimension

        :param n: The dimension of the disturbance, corresponding to the state dimension of the dynamic system
        :type n: int
        :param num_funcs: The number of basis functions (m) for each scalar dimension, defaults to 5
        :type num_funcs: int, optional
        :param decay_weights: If True, the weights of the basis functions decrease as the index of the functions increase, defaults to False
        :type decay_weights: bool, optional
        :param negative_weights: If False, the weights of the basis functions must only be positive, defaults to True
        :type negative_weights: bool, optional
        :return: A :class:`dadra.Disturbance` object
        :rtype: :class:`dadra.Disturbance`
        """
        dist_list = [
            ScalarDisturbance.poly_disturbance(
                num_funcs,
                decay_weights=decay_weights,
                negative_weights=negative_weights,
            )
            for _ in range(n)
        ]
        return cls(dist_list)

    @classmethod
    def n_zero_disturbance(cls, n):
        """Class method that creates an instance of :class:`dadra.Disturbance` consisting of zero disturbance for each dimension

        :param n: The dimension of the disturbance, corresponding to the state dimension of the dynamic system
        :type n: int
        :return: A :class:`dadra.Disturbance` object
        :rtype: :class:`dadra.Disturbance`
        """
        dist_list = [ScalarDisturbance.no_disturbance() for _ in range(n)]
        return cls(dist_list)

    def draw_alphas(self):
        """Initialize the alphas, the weights for the basis functions, for each separate scalar disturbance"""
        for dist in self.dist_list:
            dist.draw_alpha()

    def get_dist(self, n, t):
        """Computes the scalar disturbance at the specified dimension and time

        :param n: The dimension of the scalar disturbance
        :type n: int
        :param t: The time
        :type t: float
        :return: The scalar disturbance
        :rtype: float
        """
        return self.dist_list[n].d(t)



import math
import numpy as np
import time

from dadra.utils.misc_utils import format_time, normalize, normalize0
from sklearn.gaussian_process.kernels import RBF
from sklearn.preprocessing import PolynomialFeatures


def c_num_samples(epsilon, delta, n_x=2, k=10):
    """Compute the number of samples needed to satisfy the specified probabilistic guarantees for the inverse Christoffel function reachable set estimate

    :param epsilon: The accuracy parameter
    :type epsilon: float
    :param delta: The confidence parameter
    :type delta: float
    :param n_x: The state dimension, defaults to 2
    :type n_x: int
    :param k: degree of polynomial features points are mapped to
    :type k: int
    :return: The number of samples needed to satisfy the specified probabilistic guarantees
    :rtype: int
    """
    return math.ceil(
        (5 / epsilon)
        * (math.log(4 / delta) + math.comb(n_x + 2 * k, n_x) * math.log(40 / epsilon))
    )


def construct_inv_christoffel(data, d):
    """Constructs the inverse christoffel function from data

    :param data: A R^{n x p} matrix, n input points of dimension p
    :type data: numpy.ndarray
    :param d: degree of polynomial features points are mapped to
    :type d: int
    :return: C, the degree d inverse christoffel function constructed from data
    :rtype: function
    """
    # [a, b] -> [1, a, b, a^2, ab, b^2]
    poly_map = PolynomialFeatures(degree=d).fit_transform
    time1 = time.perf_counter()
    poly_mapped_data = poly_map(data)
    time2 = time.perf_counter()
    print(f"Time to apply polynomial mapping to data: {format_time(time2-time1)}")

    # scale V for numerical stability
    V = normalize0(poly_mapped_data).T
    # V is z_k(x), vector of monomials degree <= k (d)
    # M is matrix of moments
    M = V @ V.T
    time3 = time.perf_counter()
    print(f"Time to construct moment matrix: {format_time(time3-time2)}")

    # Moore-Penrose psuedo-inverse of matrix
    M_inv = np.linalg.pinv(M)
    time4 = time.perf_counter()
    print(f"Time to (pseudo)invert moment matrix: {format_time(time4-time3)}")

    def C(x):
        """The empirical inverse christoffel function for a point cloud x of i.i.d. samples

        :param x: Point cloud of i.i.d. samples, shape R^p or R^{n' x p}
        :type x: numpy.ndarray
        :return: R^{n'} vector of evaluations on data
        :rtype: numpy.ndarray
        """
        z = poly_map(x)
        if len(z.shape) == 1:
            return z @ M_inv @ z
        elif len(z.shape) == 2:
            return ((z @ M_inv) * z).sum(axis=1)

    return C


def make_rbf_kernel(scale=1.0):
    """Constructs a radical basis function kernel

    :param scale: The length scale of the kernel, defaults to 1.0
    :type scale: float
    :return: The radical basis function kernel and the diagonal of the kernel
    :rtype: tuple
    """
    rbf = RBF(scale)

    def rbf_ker(x1, x2):
        """Computes the rbf kernel from two arrays

        :param x1: Left argument of the rbf kernel, array of shape (n_samples_x1, n_features)
        :type x1: numpy.ndarray
        :param x2: Right argument of the rbf kernel, array of shape (n_samples_x2, n_features)
        :type x2: numpy.ndarray
        :return: The rbf kernel k(x1, x2), array of shape (n_sample_x1, n_samples_x2)
        :rtype: numpy.ndarray
        """
        if len(x1.shape) == 1:
            x1 = x1[None]
        if len(x2.shape) == 1:
            x2 = x2[None]
        return np.squeeze(rbf(x1, x2))

    def diag_rbf_ker(x):
        """Computes the diagonal of the kernel

        :param x: The rbf kernel
        :type x: numpy.ndarray
        :return: The diagonal of the rbf kernel
        :rtype: numpy.ndarray
        """
        return rbf.diag(x)

    return rbf_ker, diag_rbf_ker


def construct_kernelized_inv_christoffel(data, d, rho=0.0001, rbf=False, scale=1.0):
    """Constructs the kernelized inverse christoffel function from data

    :param data: R^{n x p} matrix, n input points of dimension p
    :type data: numpy.ndarray
    :param d: degree of polynomial features points are mapped to
    :type d: int
    :param rho: Constant rho, defaults to 0.0001
    :type rho: float, optional
    :param rbf: If True, a radical basis function kernel is used, defaults to False
    :type rbf: bool, optional
    :param scale: The length scale of the kernel, defaults to 1.0
    :type scale: int, optional
    :return: C, the kernelized degree d inverse christoffel function constructed from data
    :rtype: function
    """
    n = data.shape[0]
    if rbf:
        ker, diag_ker = make_rbf_kernel(scale)
    else:
        ker = lambda x1, x2: (1 + x1 @ x2.T) ** d
        diag_ker = lambda x: (1 + (x * x).sum(axis=1)) ** d

    time1 = time.perf_counter()
    X = normalize(data, 0, 1)
    VTV = ker(X, X)
    time2 = time.perf_counter()
    print(f"Time to construct data kernel matrix: {format_time(time2-time1)}")

    rho = rho * np.linalg.norm(VTV) / np.sqrt(n)
    K_inv = np.linalg.pinv(rho * np.eye(n) + VTV)
    time2 = time.perf_counter()
    print(f"Time to invert rho*I + kernel: {format_time(time2-time1)}")

    def C(x):
        """The empirical inverse christoffel function for a point cloud x of i.i.d. samples

        :param x: Point cloud of i.i.d. samples, shape R^p or R^{n' x p}
        :type x: numpy.ndarray
        :return: R^{n'} vector of evaluations on data
        :rtype: numpy.ndarray
        """
        x = normalize(x, 0, 1)
        if len(x.shape) == 1:
            kappa = ker(x, x)
            VTv = ker(X, x)
            return (kappa - VTv @ K_inv @ VTv) / rho
        elif len(x.shape) == 2:
            kappa = diag_ker(x)
            VTv = ker(X, x)
            return (kappa - (VTv * (K_inv @ VTv)).sum(axis=0)) / rho

    return C


def c_compute_contours(C, samples, grid_n=200):
    """Computes contours of C in region surrounding the input samples

    :param C: The inverse Christoffel function
    :type C: function
    :param samples: Samples from dynamical system (num_samples, n_x)
    :type samples: numpy.ndarray
    :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
    :type grid_n: int, optional
    :return: The meshgrid, and the corresponding computed contour
    :rtype: tuple
    """
    x_min = samples[:, 0].min()
    x_max = samples[:, 0].max()
    y_min = samples[:, 1].min()
    y_max = samples[:, 1].max()
    x = np.linspace(
        x_min - 0.1 * (x_max - x_min), x_max + 0.1 * (x_max - x_min), grid_n
    )
    y = np.linspace(
        y_min - 0.1 * (y_max - y_min), y_max + 0.1 * (y_max - y_min), grid_n
    )
    xv, yv = np.meshgrid(x, y)
    z = np.array([xv.flatten(), yv.flatten()]).T
    cont = C(z)
    return xv, yv, cont.reshape(grid_n, grid_n)


def c_emp_estimate(samples, C, level):
    """Computes the ratio of samples within the estimated reachable set for the Christoffel function reachable set estimation

    :param samples: Sample from dynamical system (num_samples, n_x)
    :type samples: numpy.ndarray
    :param C: The inverse Christoffel function
    :type C: function
    :param level: The maximum level value of the Christoffel function on the original sample
    :type level: float
    :return: The ratio of samples within the estimated reachability set
    :rtype: float
    """
    values = C(samples)
    num_elem = np.prod(samples.shape)
    ratio = 1 - np.where(values > level, 1, 0).sum() / num_elem
    return ratio



import numpy as np

from dadra.disturbance import Disturbance
from dadra.sampling import make_sample_n
from dadra.systems.dynamics import *
from numpy.random import default_rng
from scipy.integrate import odeint


def sample_oscillator(
    x=None, disturbance: Disturbance = None, timesteps=100, parts=1001
):
    """Obtains a sample from a Duffing Oscillator over the specified number of timesteps

    :param x: Placeholder variable over which to evaluate, defaults to None
    :type x: NoneType, optional
    :param disturbance: The disturbance to be added to each dimension, defaults to None
    :type disturbance: :class:`dadra.Disturbance`, optional
    :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
    :type timesteps: int, optional
    :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
    :type parts: int, optional
    :return: The sample from the Duffing Oscillator at the last timestep
    :rtype: numpy.ndarray
    """
    t = np.linspace(0, timesteps, parts)
    y0 = np.array([np.random.uniform(0.95, 1.05), np.random.uniform(-0.05, 0.05)])
    if disturbance is None:
        sol = odeint(duffing_oscillator, y0, t)
    else:
        disturbance.draw_alphas()
        sol = odeint(duffing_oscillator, y0, t, args=tuple([disturbance]))
    return sol[-1]


def sample_lorenz(x=None, disturbance: Disturbance = None, timesteps=100, parts=1001):
    """Obtains a sample from a Lorenz system over the specified number of timesteps

    :param x: Placeholder variable over which to evaluate, defaults to None
    :type x: NoneType, optional
    :param disturbance: The disturbance to be added to each dimension, defaults to None
    :type disturbance: :class:`dadra.Disturbance`, optional
    :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
    :type timesteps: int, optional
    :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
    :type parts: int, optional
    :return: The sample from the Lorenz system at the last timestep
    :rtype: numpy.ndarray
    """
    t = np.linspace(0, timesteps, parts)
    ru = default_rng().uniform
    z0 = np.array([ru(0, 1), ru(0, 1), ru(0, 1)])
    if disturbance is None:
        sol = odeint(lorenz_system, z0, t)
    else:
        disturbance.draw_alphas()
        sol = odeint(lorenz_system, z0, t, args=tuple([disturbance]))
    return sol[-1]


def sample_planar_quadrotor(x=None, timesteps=5, parts=5001):
    """Obtains a sample from a Planar Quadrotor Model over the specified number of timesteps

    :param x: Placeholder variable over which to evaluate, defaults to None
    :type x: NoneType, optional
    :param timesteps: The number of timesteps over which to compute the sample, defaults to 5
    :type timesteps: int, optional
    :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 5001
    :type parts: int, optional
    :return: The sample from the Planar Quadrotor Model at the last timestep
    :rtype: numpy.ndarray
    """
    g = 9.81
    K = 0.89 / 1.4
    ru = np.random.uniform
    t = np.linspace(0, timesteps, parts)

    # initial states
    y0 = np.array(
        [
            ru(-1.7, 1.7),
            ru(-0.8, 0.8),
            ru(0.3, 2.0),
            ru(-1.0, 1.0),
            ru(-np.pi / 12, np.pi / 12),
            ru(-np.pi / 2, np.pi / 2),
        ]
    )
    sol = odeint(
        planar_quadrotor,
        y0,
        t,
        args=(ru(-1.5 + g / K, 1.5 + g / K), ru(-np.pi / 4, np.pi / 4)),
    )
    return sol[-1]


def sample_traffic(x=None, nx=6, timesteps=100, parts=10001):
    """Obtains a sample from a Monotone Traffic Model over the specified number of timesteps

    :param x: Placeholder variable over which to evaluate, defaults to None
    :type x: NoneType, optional
    :param nx: The state dimension of the model, defaults to 6
    :type nx: int, optional
    :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
    :type timesteps: int, optional
    :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 10001
    :type parts: int, optional
    :return: The sample from the Monotone Traffic Model at the last timestep
    :rtype: numpy.ndarray
    """
    t = np.linspace(0, timesteps, parts)
    y0 = 100 * np.ones(nx) + 100 * np.random.rand(nx)
    sol = odeint(traffic, y0, t)
    return sol[-1]


def sample_quadrotor(x=None, timesteps=5, parts=5001):
    """Obtains sample from the 12-state Quadrotor Model across all timesteps

    :param x: Placeholder variable over which to evaluate, defaults to None
    :type x: NoneType, optional
    :param timesteps: The number of timesteps over which to compute the sample, defaults to 5
    :type timesteps: int, optional
    :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 5001
    :return: The sample from the 12-state Quadrotor Model accross all timesteps
    :rtype: numpy.ndarray
    """
    t = np.linspace(0, timesteps, parts)
    ru = default_rng().uniform
    x1_0, x2_0, x3_0, x4_0, x5_0, x6_0 = [ru(-0.4, 0.4) for _ in range(6)]
    rand_states = np.array([x1_0, x2_0, x3_0, x4_0, x5_0, x6_0])
    x_0 = np.concatenate((rand_states, np.zeros(6)))
    sol = odeint(quadrotor, x_0, t)
    return sol


sample_oscillator_n = make_sample_n(sample_oscillator)
"""Function to sample from a Duffing Oscillator

:param n: The number of samples to compute.
:type n: int
"""


sample_lorenz_n = make_sample_n(sample_lorenz)
"""Function to sample from a Lorenz system

:param n: The number of samples to compute.
:type n: int
"""

sample_planar_quadrotor_n = make_sample_n(sample_planar_quadrotor)
"""Function to sample from a Planar Quadrotor Model

:param n: The number of samples to compute.
:type n: int
"""


sample_traffic_n = make_sample_n(sample_traffic)
"""Function to sample from a Monotone Traffic Model

:param n: The number of samples to compute.
:type n: int
"""

sample_quadrotor_n = make_sample_n(sample_quadrotor)
"""Function to sample from a 12-state Quadrotor Model

:param n: The number of samples to compute.
:type n: int
"""



import cvxpy as cp
import math
import numpy as np

from collections import OrderedDict
from functools import partial
from multiprocessing import Pool, cpu_count
from scipy.optimize import minimize_scalar
from tqdm.auto import tqdm


def p_num_samples(epsilon, delta, n_x=3, const=None):
    """Compute the number of samples needed to satisfy the specified probabilistic guarantees for the p-norm ball reachable set estimate

    :param epsilon: The accuracy parameter
    :type epsilon: float
    :param delta: The confidence parameter
    :type delta: float
    :param n_x: The state dimension, defaults to 3
    :type n_x: int
    :param const: The constraints placed on the parameters A and b, defaults to None
    :type const: string, optional
    :return: The number of samples needed to satisfy the specified probabilistic guarantees
    :rtype: int
    """
    if const is None:
        n_theta = 0.5 * (n_x ** 2 + 3 * n_x)
    elif const == "diagonal":
        n_theta = 2 * n_x
    elif const == "scalar":
        n_theta = 1
    N = math.ceil(math.e * (math.log(1 / delta) + n_theta) / (epsilon * (math.e - 1)))
    return N


def solve_p_norm(sample, n_x=3, p=2, const=None):
    """Solves the scenario relaxation problem for the given sample with p-Norm Balls

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param const: The constraints placed on the parameters A and b, defaults to None
    :type const: string, optional
    :return: The values of matrix A and vector b corresponding to the optimal p-Norm Ball, as well as the status of the optimizer.
    :rtype: tuple
    """
    if const is None:
        A = cp.Variable((n_x, n_x), symmetric=True)
        b = cp.Variable((n_x, 1))
    elif const == "diagonal":
        a = cp.Variable((n_x, 1))
        A = cp.diag(a)
        b = cp.Variable((n_x, 1))
    elif const == "scalar":
        sigma = cp.Variable()
        A = sigma * np.identity(n_x)
        b = np.zeros((n_x, 1))

    obj = cp.Minimize(-cp.log_det(A))
    constraints = [cp.pnorm(A @ r.reshape(n_x, 1) - b, p=p) <= 1 for r in sample]
    prob = cp.Problem(obj, constraints)
    prob.solve()

    if const != "scalar":
        return A.value, b.value, prob.status
    else:
        return A, b, prob.status


def multi_p_norm(samples, p=2, const=None):
    """Computes a the p-norm ball reachable set estimates across a series of timesteps

    :param samples: The samples from a dynamic system across time, an array of shape (num_samples, timesteps, state_dim)
    :type samples: numpy.ndarray
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param const: The constraints placed on the parameters A and b, defaults to None
    :type const: string, optional
    :raises ValueError: [description]
    :return: [description]
    :rtype: [type]
    """
    if len(samples.shape) != 3:
        raise ValueError("Samples must be of shape (num_samples, timesteps, state_dim")
    n_x = samples.shape[2]
    keys = ("A", "b", "status")
    solutions = [
        dict(zip(keys, solve_p_norm(sample, n_x, p, const)))
        for sample in tqdm(samples.swapaxes(0, 1))
    ]

    return solutions


def p_norm_cont(arr, axis, default_val, n_x, A_val, b_val, p, minimum=True):
    """Solve for the optimal value that satisfies the p-Norm Ball conditions at the specified axis

    :param arr: Array of shape (n_x - 1,) containing the independent variables of the p-norm condition
    :type arr: numpy.ndarray
    :param axis: The axis of the dependent variable for which to solve for (i.e. z -> axis=2).
    :type axis: int
    :param default_val: The value to return if no solution for the dependent variable is found that satisfies the p-norm conditions
    :type default_val: float
    :param n_x: The state dimension
    :type n_x: int
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param p: The order of p-norm
    :type p: int
    :param minimum: True if optimizing for the minimal value of the dependent variable that satisfies the p-norm conditions, defaults to True
    :type minimum: bool, optional
    :return: The value at the specified axis which corresponds the the optimal value of the (n_x, 1) vector that satisfies the p-Norm Ball conditions at the specified axis
    :rtype: float
    """
    vec = cp.Variable((n_x, 1))
    other_dims = list(range(n_x))
    other_dims.remove(axis)
    constraints = [vec[i][0] == arr[j] for i, j in zip(other_dims, range(n_x - 1))]
    constraints.append(cp.pnorm(A_val @ vec - b_val, p=p) <= 1)
    if minimum:
        obj = cp.Minimize(vec[axis])
    else:
        obj = cp.Maximize(vec[axis])
    prob = cp.Problem(obj, constraints)

    try:
        prob.solve()
    except:
        return default_val

    if prob.status != "optimal":
        return default_val

    return vec.value[axis]


def p_norm_cont_proj(arr, axis, default_val, n_x, A_val, b_val, p):
    """Minimizes the p-Norm value with respect to value at the specified axis.

    :param arr: Array of shape (n_x - 1,) containing the independent variables of the p-norm condition.
    :type arr: numpy.ndarray
    :param axis: The axis of the dependent variable for which to solve for (i.e. z -> axis=2).
    :type axis: int
    :param default_val: The value to return if no solution for the dependent variable is found that satisfies the p-norm conditions.
    :type default_val: float
    :param n_x: The state dimension.
    :type n_x: int
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball.
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball.
    :type b_val: numpy.ndarray
    :param p: The order of p-norm.
    :type p: int
    :return: The value at the specified axis which corresponds the the minimum p-Norm value of the (n_x, 1)  vector.
    :rtype: float
    """
    vec = np.zeros((n_x))
    other_dims = list(range(n_x))
    other_dims.remove(axis)
    for i, j in zip(other_dims, range(n_x - 1)):
        vec[i] = arr[j]

    def f(x):
        vec[axis] = x
        return np.linalg.norm(A_val @ vec.reshape((n_x, 1)) - b_val, ord=p)

    res = minimize_scalar(f)
    vec[axis] = res.x

    if np.linalg.norm(A_val @ vec.reshape((n_x, 1)) - b_val, ord=p) <= 1:
        return res.x
    else:
        return default_val


def p_compute_contour_2D(sample, A_val, b_val, cont_axis=2, n_x=3, p=2, grid_n=200):
    """Computes the 3D contour for 2 dimensions based on sample data and the A_val, and b_val corresponding to the optimal p-norm ball.

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param cont_axis: The axis for which the contours are to be solved for, defaults to 2
    :type cont_axis: int, optional
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
    :type grid_n: int, optional
    :return: The meshgrid, corresponding computed contour, and the extremum values for the chosen axis
    :rtype: tuple
    """
    x_min, x_max = sample[:, 0].min(), sample[:, 0].max()
    y_min, y_max = sample[:, 1].min(), sample[:, 1].max()
    z_min, z_max = sample[:, 2].min(), sample[:, 2].max()

    x = np.linspace(
        x_min - 0.4 * (x_max - x_min), x_max + 0.4 * (x_max - x_min), grid_n
    )
    y = np.linspace(
        y_min - 0.4 * (y_max - y_min), y_max + 0.4 * (y_max - y_min), grid_n
    )
    z = np.linspace(
        x_min - 0.4 * (z_max - z_min), z_max + 0.4 * (z_max - z_min), grid_n
    )

    if cont_axis == 2:
        d0, d1 = np.meshgrid(x, y)
        c_min, c_max = z_min, z_max
    elif cont_axis == 1:
        d0, d1 = np.meshgrid(x, z)
        c_min, c_max = y_min, y_max
    elif cont_axis == 0:
        d0, d1 = np.meshgrid(y, z)
        c_min, c_max = x_min, x_max

    d2 = np.array([d0.flatten(), d1.flatten()]).T

    solve_cont_d2 = partial(
        p_norm_cont_proj,
        axis=cont_axis,
        default_val=c_max + 1,
        n_x=n_x,
        A_val=A_val,
        b_val=b_val,
        p=p,
    )
    cont = np.fromiter(map(solve_cont_d2, d2), dtype=np.float64).reshape(grid_n, grid_n)

    return d0, d1, cont, c_min, c_max


def p_compute_contour_3D(
    sample, A_val, b_val, cont_axis=2, n_x=3, p=2, grid_n=200, stretch=0.4
):
    """Computes the 3D contour for 3 dimensions based on sample data and the A_val, and b_val corresponding to the optimal p-norm ball.

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param cont_axis: The axis for which the contours are to be solved for, defaults to 2
    :type cont_axis: int, optional
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
    :type grid_n: int, optional
    :param stretch: The factor by which to stretch the grid used to compute the contour, defaults to 0.4
    :type stretch: float, optional
    :param minimum: True if optimizing for the minimal value of the dependent variable that satisfies the p-norm conditions, defaults to True
    :type minimum: bool, optional
    :return: The meshgrid, corresponding computed contour, and the extremum values for the chosen axis
    :rtype: tuple
    """
    x_min, x_max = sample[:, 0].min(), sample[:, 0].max()
    y_min, y_max = sample[:, 1].min(), sample[:, 1].max()
    z_min, z_max = sample[:, 2].min(), sample[:, 2].max()

    x = np.linspace(
        x_min - stretch * (x_max - x_min), x_max + stretch * (x_max - x_min), grid_n
    )
    y = np.linspace(
        y_min - stretch * (y_max - y_min), y_max + stretch * (y_max - y_min), grid_n
    )
    z = np.linspace(
        x_min - stretch * (z_max - z_min), z_max + stretch * (z_max - z_min), grid_n
    )

    if cont_axis == 2:
        d0, d1 = np.meshgrid(x, y)
        c_min, c_max = z_min, z_max
    elif cont_axis == 1:
        d0, d1 = np.meshgrid(x, z)
        c_min, c_max = y_min, y_max
    elif cont_axis == 0:
        d0, d1 = np.meshgrid(y, z)
        c_min, c_max = x_min, x_max

    d2 = np.array([d0.flatten(), d1.flatten()]).T

    solve_cont_d2_min = partial(
        p_norm_cont,
        axis=cont_axis,
        default_val=c_max + 1,
        n_x=n_x,
        A_val=A_val,
        b_val=b_val,
        p=p,
        minimum=True,
    )

    solve_cont_d2_max = partial(
        p_norm_cont,
        axis=cont_axis,
        default_val=c_min - 1,
        n_x=n_x,
        A_val=A_val,
        b_val=b_val,
        p=p,
        minimum=False,
    )

    cont_min = np.fromiter(map(solve_cont_d2_min, d2), dtype=np.float64).reshape(
        grid_n, grid_n
    )
    cont_max = np.fromiter(map(solve_cont_d2_max, d2), dtype=np.float64).reshape(
        grid_n, grid_n
    )

    return d0, d1, cont_min, cont_max, c_min, c_max


def p_compute_vals(sample, A_val, b_val, p=2, grid_n=200):
    """Computes the values within a p-norm ball in 1 dimension

    :param sample: The sample from a specific time step, an array of shape (num_samples,)
    :type sample: numpy.ndarray
    :param A_val: The matrix of shape (1, 1) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (1, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param grid_n: The number of points to test for the p-norm ball estimation at each a given time step, defaults to 200
    :type grid_n: int, optional
    :return: The values within the p-norm ball
    :rtype: list
    """
    # assuming sample is (num_samples,) shaped array
    y_min, y_max = sample.min(), sample.max()
    y = np.linspace(
        y_min - 0.4 * (y_max - y_min), y_max + 0.4 * (y_max - y_min), grid_n
    )

    vals = []
    for v in y:
        if np.linalg.norm(A_val @ np.array([[v]]) - b_val, ord=p) <= 1:
            vals.append(v)
    return vals


def p_get_dict(i, samples, solution_list, items, grid_n=50):
    """Generates a dictionary where keys are the passed in labels and values are the output of p_compute_contour_3D, the contour information for a p-norm ball reachable set estimate at a given time

    :param i: The index
    :type i: int
    :param samples: Array of shape (num_samples, time, 3)
    :type samples: numpy.ndarray
    :param solution_list: List of solutions where each solution is a dictionary with keys ("A", "b", "status"), defaults to None
    :type solution_list: list, optional
    :param items: A list of specific indices corresponding to the times at which the p-norm ball reachable set estimate will be chosen
    :type items: list
    :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 50
    :type grid_n: int, optional
    :return: A dictionary where keys are the passed in labels and values are the output of p_compute_contour_3D, the contour information for a p-norm ball reachable set estimate at a given time
    :rtype: dict
    """
    labels = ["xv", "yv", "z_cont", "z_cont2", "z_min", "z_max"]
    index = items[i]
    A_i, b_i = solution_list[index]["A"], solution_list[index]["b"]
    return dict(
        zip(labels, p_compute_contour_3D(samples[:, index, :], A_i, b_i, grid_n=grid_n))
    )


def p_dict_list(
    samples, solution_list, num_parts, num_indices=20, logspace=True, grid_n=50
):
    """Generates a list of dictionaries, each containing the contour information for a p-norm ball reachable set estimate at a given time

    :param samples: Array of shape (num_samples, time, 3)
    :type samples: numpy.ndarray
    :param solution_list: List of solutions where each solution is a dictionary with keys ("A", "b", "status"), defaults to None
    :type solution_list: list, optional
    :param num_parts: The number of total timesteps for the samples
    :type num_parts: int
    :param num_indices: The number of indices corresponding to the number of reachable set estimates made at different times, defaults to 20
    :type num_indices: int, optional
    :param logspace: If True, a logarithmic scale is used for choosing times to compute reachable set estimates, defaults to True
    :type logspace: bool, optional
    :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 50
    :type grid_n: int, optional
    :return: A list of dictionaries, each containing the contour information for a p-norm ball reachable set estimate at a given time
    :rtype: list
    """
    if logspace:
        log_ceil = np.log10(num_parts)
        items = list(
            OrderedDict.fromkeys(
                [int(i) - 1 for i in np.logspace(0, log_ceil, num_indices)]
            )
        )
    else:
        items = [int(i) for i in np.linspace(0, num_parts, num_indices)]

    get_dict = partial(
        p_get_dict,
        samples=samples,
        solution_list=solution_list,
        items=items,
        grid_n=grid_n,
    )

    p = Pool(cpu_count())
    dict_list = [
        d for d in tqdm(p.imap(get_dict, np.arange(len(items))), total=len(items))
    ]

    return dict_list


def p_emp_estimate(samples, A_val, b_val, n_x=3, p=2):
    """Computes the ratio of samples within the estimated reachable set for the p-norm ball reachable set estimation

    :param samples: Sample from dynamical system (num_samples, n_x)
    :type samples: numpy.ndarray
    :param A_val: The matrix of shape (n_x, n_x) corresponding to the optimal p-norm ball
    :type A_val: numpy.ndarray
    :param b_val: The vector of shape (n_x, 1) corresponding to the optimal p-norm ball
    :type b_val: numpy.ndarray
    :param n_x: The state dimension, defaults to 3
    :type n_x: int, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :return: The ratio of samples within the estimated reachability set
    :rtype: float
    """
    num_samples = samples.shape[0]
    count = 0
    for sample in samples:
        vec = sample.reshape(n_x, 1)
        if np.linalg.norm(A_val @ vec - b_val, ord=p) <= 1:
            count += 1
    return count / num_samples



import matplotlib.animation as animation
import matplotlib.pyplot as plt
import numpy as np


def plot_sample(sample, fig_name):
    """Plots samples in 2D

    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param fig_name: The name of the file to save the plot to
    :type fig_name: string
    """
    if sample.shape[1] == 2:
        fig, axs = plt.subplots(1, figsize=(5, 5))
        axs.plot(sample[:, 0], sample[:, 1], "k.")
        axs.set_xlabel("x")
        axs.set_ylabel("y")
        axs.set_title("contour in x-y plane")
    else:
        fig, axs = plt.subplots(3, figsize=(5, 15))
        axs[0].plot(sample[:, 0], sample[:, 1], "k.")
        axs[0].set_xlabel("x")
        axs[0].set_ylabel("y")
        axs[1].plot(sample[:, 0], sample[:, 2], "k.")
        axs[1].set_xlabel("x")
        axs[1].set_ylabel("z")
        axs[2].plot(sample[:, 1], sample[:, 2], "k.")
        axs[2].set_xlabel("y")
        axs[2].set_ylabel("z")
        plt.savefig(fig_name, bbox_inches="tight", facecolor="white")


def plot_sample_time(
    time_x, samples, fig_name, figsize=(10, 10), color="default", **kwargs
):
    """Plots trajectory of samples over time.

    :param time_x: The time corresponding to the independent variable
    :type time_x: numpy.array
    :param samples: The samples of shape (num_samples, parts)
    :type samples: numpy.array
    :param fig_name: The name of the file to save the plot to
    :type fig_name: str
    :param figsize: the size of the figure to be saved, defaults to (10, 10)
    :type figsize: tuple, optional
    :param color: The color to draw the lines in, defaults to "default"
    :type color: str, optional
    :raises ValueError: If samples is not an array of shape (num_samples, parts)
    """
    if len(samples.shape) != 2:
        raise ValueError("Samples must be of shape (num_samples, parts)")

    fig, axs = plt.subplots(1, figsize=figsize)
    for i in range(samples.shape[0]):
        if color == "default":
            axs.plot(time_x, samples[i])
        else:
            axs.plot(time_x, samples[i], color=color)

    xmin, xmax, ymin, ymax = plt.axis()

    for key in kwargs:
        if key == "x":
            for val in kwargs["x"]:
                axs.plot([val, val], [ymin, ymax], "k", alpha=0.5)
        elif key == "y":
            for val in kwargs["y"]:
                axs.plot([xmin, xmax], [val, val], "k", alpha=0.5)
        elif key == "title":
            axs.set_title(kwargs["title"])
        elif key == "ylabel":
            axs.set_ylabel(kwargs["ylabel"])

    axs.set_xlabel("time")
    plt.savefig(fig_name, bbox_inches="tight", facecolor="white")


def plot_reach_time(
    time_x,
    samples,
    min_vals,
    max_vals,
    fig_name,
    num_samples_show=0,
    figsize=(10, 10),
    c1="palevioletred",
    c2="lavenderblush",
    **kwargs,
):
    """Plots the reachable set estimate (as well as possibly some samples) over time.

    :param time_x: The time corresponding to the independent variable
    :type time_x: numpy.array
    :param samples: The samples of shape (num_samples, parts)
    :type samples: numpy.array
    :param min_vals: The lower bound of the reachable set estimate
    :type min_vals: list
    :param max_vals: The upper bound of the reachable set estimate
    :type max_vals: list
    :param fig_name: The name of the file to save the plot to
    :type fig_name: str
    :param num_samples_show: The number of sample trajectories to show in addition to the reachable set, defaults to 0
    :type num_samples_show: int, optional
    :param figsize: the size of the figure to be saved, defaults to (10, 10)
    :type figsize: tuple, optional
    :param c1: Color 1, used for the reachable set estimate, defaults to "palevioletred"
    :type c1: str, optional
    :param c2: Color 2, used for the samples, defaults to "lavendarblush"
    :type c2: str, optional
    """
    fig, axs = plt.subplots(1, figsize=figsize)

    axs.plot(time_x, min_vals, c=c1)
    axs.plot(time_x, max_vals, c=c1)
    axs.fill_between(time_x, min_vals, max_vals, color=c1)

    for i in range(num_samples_show):
        axs.plot(time_x, samples[i], color=c2, alpha=0.2)

    xmin, xmax, ymin, ymax = plt.axis()

    # assume kwargs are of the form {"x"=[1], "y"=[0.98, 1.4]}
    for key in kwargs:
        if key == "x":
            for val in kwargs["x"]:
                axs.plot([val, val], [ymin, ymax], "k", alpha=0.5)
        elif key == "y":
            for val in kwargs["y"]:
                axs.plot([xmin, xmax], [val, val], "k", alpha=0.5)
        elif key == "title":
            axs.set_title(kwargs["title"])
        elif key == "ylabel":
            axs.set_ylabel(kwargs["ylabel"])

    axs.set_xlabel("time")
    plt.savefig(fig_name, bbox_inches="tight", facecolor="white")


def grow_plot_3d(samples, fig_name, gif_name, figsize, **kwargs):
    """Creates a gif of a three dimensional trajectory that is animated to grow over time

    :param samples: Array of shape (num_samples, time, 3)
    :type samples: numpy.ndarray
    :param fig_name: The name of the static figure to be saved
    :type fig_name: str
    :param gif_name: The name of the animated gif to be saved, defaults to None
    :type gif_name: str, optional
    :raises ValueError: If the samples are not of shape (num_samples, time, 3)
    """
    if len(samples.shape) != 3 or samples.shape[2] != 3:
        raise ValueError("The shape of samples must be (num_samples, time, 3)")

    fig = plt.figure(figsize=figsize)
    ax = plt.axes(projection="3d")

    lines = []
    for sample in samples:
        (line,) = ax.plot(sample[:, 0], sample[:, 1], sample[:, 2], alpha=0.8)
        lines.append(line)

    if gif_name is not None:
        ax.set_xlabel(kwargs.get("xlabel", "x"))
        ax.set_ylabel(kwargs.get("ylabel", "y"))
        ax.set_zlabel(kwargs.get("zlabel", "z"))
        ax.set_title(kwargs.get("title", "trajectory over time"))

        def update(num, samples, lines):
            """Inner function to update line plot in 3D over time

            :param num: The number of frames, and the number of points to plot
            :type num: int
            :param samples: Array of shape (num_samples, time, 3)
            :type samples: numpy.ndarray
            :param lines: A list of Line3D objects
            :type lines: list
            :return: An updated list of Line3D objects
            :rtype: list
            """
            for line, sample in zip(lines, samples):
                line.set_xdata(sample[:num, 0])
                line.set_ydata(sample[:num, 1])
                line.set_3d_properties(sample[:num, 2])
            ax.view_init(azim=num)
            return lines

        ani = animation.FuncAnimation(
            fig,
            update,
            frames=np.arange(0, 362, 2),
            fargs=[samples, lines],
            interval=100,
        )

        ani.save(gif_name)

    plt.savefig(fig_name)


def plot_contour_2D(
    xv1,
    yv1,
    z_cont,
    z_level,
    sample,
    fig_name,
    xv2=None,
    zv1=None,
    y_cont=None,
    y_level=None,
    yv2=None,
    zv2=None,
    x_cont=None,
    x_level=None,
):
    """Plots the contours provided (output of utils.compute_contour) in 2D

    :param xv1: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type xv1: numpy.ndarray
    :param yv1: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type yv1: numpy.ndarray
    :param z_cont: A (grid_n, grid_n) matrix, each point a function of xv1 and yv1
    :type z_cont: numpy.ndarray
    :param z_level: The z-value at which the level set is to be drawn
    :type z_level: float
    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param fig_name: The name of the file to save the plot to
    :type fig_name: string
    :param xv2: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type xv2: numpy.ndarray, optional
    :param zv1: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type zv1: numpy.ndarray, optional
    :param y_cont: A (grid_n, grid_n) matrix, each point a function of xv2 and zv1, defaults to None
    :type y_cont: numpy.ndarray, optional
    :param y_level: The y-value at which the level set is to be drawn, defaults to None
    :type y_level: float, optional
    :param yv2: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type yv2: numpy.ndarray, optional
    :param zv2: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension, defaults to None
    :type zv2: numpy.ndarray, optional
    :param x_cont: A (grid_n, grid_n) matrix, each point a function of yv2 and zv2, defaults to None
    :type x_cont: numpy.ndarray, optional
    :param x_level: The x-value at which the level set is to be drawn, defaults to None
    :type x_level: float, optional
    """
    if y_cont is not None and x_cont is not None:
        fig, axs = plt.subplots(3, figsize=(5, 15))

    elif y_cont is not None:
        fig, axs = plt.subplots(2, figsize=(5, 10))

    else:
        fig, axs = plt.subplots(1, figsize=(5, 5))
        axs.plot(sample[:, 0], sample[:, 1], "k.")
        axs.contour(xv1, yv1, z_cont, levels=[z_level], colors="b")
        axs.set_xlabel("x")
        axs.set_ylabel("y")
        axs.set_title("samples in x-y plane")
        plt.show()
        plt.savefig(fig_name, bbox_inches="tight", facecolor="white")
        return

    axs[0].plot(sample[:, 0], sample[:, 1], "k.")
    axs[0].contour(xv1, yv1, z_cont, levels=[z_level], colors="b")
    axs[0].set_xlabel("x")
    axs[0].set_ylabel("y")
    axs[0].set_title("samples in x-y plane")

    if y_cont is not None:
        axs[1].plot(sample[:, 0], sample[:, 2], "k.")
        axs[1].contour(xv2, zv1, y_cont, levels=[y_level], colors="b")
        axs[1].set_xlabel("x")
        axs[1].set_ylabel("z")
        axs[1].set_title("samples in x-z plane")

        if x_cont is not None:
            axs[2].plot(sample[:, 1], sample[:, 2], "k.")
            axs[2].contour(yv2, zv2, x_cont, levels=[x_level], colors="b")
            axs[2].set_xlabel("y")
            axs[2].set_ylabel("z")
            axs[2].set_title("samples in y-z plane")

    plt.savefig(fig_name, bbox_inches="tight", facecolor="white")


def plot_contour_3D(
    xv,
    yv,
    z_cont,
    z_min,
    z_max,
    sample,
    fig_name,
    gif_name=None,
    z_cont2=None,
):
    """Plots the contours in 3D with the option for saving an animated gif of the rotating graph

    :param xv: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type xv: numpy.ndarray
    :param yv: A (grid_n, grid_n) matrix with the elements of a (grid_n, 1) vector repeated along the first dimension
    :type yv: numpy.ndarray
    :param z_cont: A (grid_n, grid_n) matrix, each point a function of xv1 and yv1 (corresponding to the maximum optimal solutions that satisfy the p-norm)
    :type z_cont: numpy.ndarray
    :param z_min: The minimum z-value
    :type z_min: float
    :param z_max: The maximum z-value
    :type z_max: float
    :param sample: Sample from dynamical system (num_samples, n_x)
    :type sample: numpy.ndarray
    :param fig_name: The name of the file to save the plot to
    :type fig_name: string
    :param gif_name: The name of the file to save the gif to, defaults to None
    :type gif_name: string, optional
    :param z_cont2: A (grid_n, grid_n) matrix, each point a function of xv and yv (corresponding to the minimum optimal solutions that satisfy the p-norm), defaults to None
    :type z_cont2: numpy.ndarray, optional
    """
    fig = plt.figure(figsize=(10, 10))
    ax = plt.axes(projection="3d")
    v_min = z_min - 0.25 * (z_max - z_min)
    v_max = z_max + 0.25 * (z_max - z_min)
    ax.scatter3D(
        sample[:, 0],
        sample[:, 1],
        sample[:, 2],
        c=sample[:, 2],
        cmap="binary",
        vmin=v_min,
        vmax=z_max,
    )
    if z_cont2 is None:
        ax.contour3D(
            xv, yv, z_cont, 50, cmap="Blues", vmin=v_min, vmax=v_max, alpha=0.8
        )
    else:
        abs_min = np.min(z_cont)
        abs_max = np.max(z_cont2)
        abs_avg = (abs_min + abs_max) / 2
        levels_1 = list(np.linspace(abs_min, abs_avg, 25))
        levels_2 = list(np.linspace(abs_avg, abs_max, 25))
        ax.contour3D(
            xv,
            yv,
            z_cont,
            50,
            cmap="Blues",
            vmin=v_min,
            vmax=v_max,
            levels=levels_1,
        )
        ax.contour3D(
            xv,
            yv,
            z_cont2,
            50,
            cmap="Blues",
            vmin=v_min,
            vmax=v_max,
            levels=levels_2,
        )

    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")
    ax.set_title("3D contour")

    if gif_name is not None:

        def rotate(angle):
            ax.view_init(azim=angle)

        rot_animation = animation.FuncAnimation(
            fig, rotate, frames=np.arange(0, 362, 2), interval=100
        )
        rot_animation.save(gif_name, dpi=100)

    plt.savefig(fig_name, bbox_inches="tight", facecolor="white")


def plot_contour_3D_time(
    samples, dict_list, fig_name, gif_name=None, figsize=(10, 10), **kwargs
):
    """Plots the contours of the p-norm reachable set estimate along with the samples over time

    :param samples: Array of shape (num_samples, time, 3)
    :type samples: numpy.ndarray
    :param dict_list: A list of dictionaries, each containing the contour information for a p-norm ball reachable set estimate at a given time
    :type dict_list: list
    :param fig_name: The name of the file to save the plot to
    :type fig_name: string
    :param gif_name: The name of the file to save the gif to, defaults to None
    :type gif_name: string, optional
    :param figsize: The size of the figure to be saved, defaults to (10, 10)
    :type figsize: tuple, optional
    """
    fig = plt.figure(figsize=figsize)
    ax = plt.axes(projection="3d")

    ax.set_xlabel(kwargs.get("xlabel", "x"))
    ax.set_ylabel(kwargs.get("ylabel", "y"))
    ax.set_zlabel(kwargs.get("zlabel", "z"))
    ax.set_title(kwargs.get("title", "reachable set over time"))

    for sample in samples:
        ax.plot(sample[:, 0], sample[:, 1], sample[:, 2], alpha=0.8)

    for curr_dict in dict_list:
        xv = curr_dict["xv"]
        yv = curr_dict["yv"]
        z_cont = curr_dict["z_cont"]
        z_min = curr_dict["z_min"]
        z_max = curr_dict["z_max"]
        z_cont2 = curr_dict["z_cont2"]

        v_min = z_min - 0.25 * (z_max - z_min)
        v_max = z_max + 0.4 * (z_max - z_min)

        abs_min = np.min(z_cont)
        abs_max = np.max(z_cont2)
        abs_avg = (abs_min + abs_max) / 2
        levels_1 = list(np.linspace(abs_min, abs_avg, 25))
        levels_2 = list(np.linspace(abs_avg, abs_max, 25))

        ax.contour3D(
            xv,
            yv,
            z_cont,
            50,
            cmap="Blues",
            vmin=v_min,
            vmax=v_max,
            levels=levels_1,
            alpha=0.1,
        )

        ax.contour3D(
            xv,
            yv,
            z_cont2,
            50,
            cmap="Blues",
            vmin=v_min,
            vmax=v_max,
            levels=levels_2,
            alpha=0.1,
        )

    if gif_name is not None:

        def rotate(num):
            ax.view_init(elev=20, azim=num)

        ani = animation.FuncAnimation(
            fig, rotate, frames=np.arange(0, 362, 2), interval=50
        )

        ani.save(gif_name)

    plt.savefig(fig_name, bbox_inches="tight", facecolor="white")



from functools import partial
import numpy as np
import time
import warnings

from dadra.dyn_sys import Sampler, SimpleSystem, System
from dadra.utils.graph_utils import (
    grow_plot_3d,
    plot_contour_2D,
    plot_contour_3D,
    plot_contour_3D_time,
    plot_reach_time,
    plot_sample,
    plot_sample_time,
)
from dadra.utils.christoffel_utils import (
    c_compute_contours,
    c_emp_estimate,
    c_num_samples,
    construct_inv_christoffel,
    construct_kernelized_inv_christoffel,
)
from dadra.utils.misc_utils import format_time
from dadra.utils.p_utils import (
    multi_p_norm,
    p_compute_contour_2D,
    p_compute_contour_3D,
    p_compute_vals,
    p_dict_list,
    p_emp_estimate,
    p_get_dict,
    p_num_samples,
    solve_p_norm,
)


class Estimator:
    """Class which allows for estimation of the reachable sets for a given dynamical system.

    :param dyn_sys: An instance of :class:`dadra.System`
    :type dyn_sys: System
    :param epsilon: The accuracy parameter, defaults to 0.05
    :type epsilon: float, optional
    :param delta: The confidence parameter, defaults to 1e-9
    :type delta: float, optional
    :param christoffel: If True, uses Christoffel functions to estimate the reachable set. If False, the p-norm method is used, defaults to False
    :type christoffel: bool, optional
    :param p: The order of p-norm, defaults to 2
    :type p: int, optional
    :param const: The constraints placed on the parameters A and b, defaults to None
    :type const: str, optional
    :param normalize: If true, the sample is normalized, defaults to True
    :type normalize: bool, optional
    :param d: The degree of polynomial features points are mapped to, defaults to 10
    :type d: int, optional
    :param kernelized: If True, the kernelized inverse christoffel function is constructed, defaults to False
    :type kernelized: bool, optional
    :param rho: Inverse Christoffel function constant rho, defaults to 0.0001
    :type rho: float, optional
    :param rbf: If True, a radical basis function kernel is used, defaults to False
    :type rbf: bool, optional
    :param scale: The length scale of the kernel, defaults to 1.0
    :type scale: float, optional
    :param iso_dim: Isolates the samples at the specified dimensions, currently only implemented when dyn_sys.all_time is True, defaults to None
    :type iso_dim: list, optional
    """

    def __init__(
        self,
        dyn_sys: System,
        epsilon=0.05,
        delta=1e-9,
        christoffel=False,
        p=2,
        const=None,
        normalize=True,
        d=10,
        kernelized=False,
        rho=0.0001,
        rbf=False,
        scale=1.0,
        iso_dim=None,
    ):

        self.dyn_sys = dyn_sys
        self.check_sys()

        self.epsilon = epsilon
        self.delta = delta
        self.christoffel = christoffel

        if not christoffel:
            self.p = p
            self.const = const
        else:
            self.const = const
            self.d = d
            self.kernelized = kernelized
            self.rho = rho
            self.rbf = rbf
            self.scale = scale

        self.normalize = normalize
        self.num_samples = self.get_num_samples()

        self.samples = None
        self.iso_dim = None

        if not christoffel:
            if not self.dyn_sys.all_time:
                self.A = None
                self.b = None
                self.status = None
            else:
                self.iso_dim = (
                    iso_dim
                    if isinstance(iso_dim, list) or iso_dim is None
                    else [iso_dim]
                )
                self.num_iso_samples = self.get_num_samples(iso=True)
                self.solution_list = None
                self.num_opt = None
                self.num_opt_in = None
                self.num_fail = None
        else:
            self.C = None
            self.level = None

        if iso_dim is not None:
            self.iso_samples = None

    @classmethod
    def estimator_from_sample_func(
        cls,
        sample_fn,
        state_dim,
        timesteps,
        parts,
        all_time=False,
        epsilon=0.05,
        delta=1e-9,
        christoffel=False,
        p=2,
        const=None,
        normalize=True,
        d=10,
        kernelized=False,
        rho=0.0001,
        rbf=False,
        scale=0.5,
        iso_dim=None,
    ):
        """Class method that allows for an instance of :class:`dadra.Estimator` to be initialized using a sampling function, and the components for an instance of :class:`Sampler` rather than explicitly passing in a :class:`System` object

        :param sample_fn: A function to sample from
        :type sample_fn: function
        :param state_dim: The degrees of freedom of the system
        :type state_dim: int
        :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
        :type timesteps: int, optional
        :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
        :type parts: int, optional
        :param all_time: If True, each sample will include all timesteps from the system, rather than only the last timestep, defaults to False
        :type all_time: bool, optional
        :param epsilon: The accuracy parameter, defaults to 0.05
        :type epsilon: float, optional
        :param delta: The confidence parameter, defaults to 1e-9
        :type delta: float, optional
        :param christoffel: If True, uses Christoffel functions to estimate the reachable set. If False, the p-norm method is used, defaults to False
        :type christoffel: bool, optional
        :param p: The order of p-norm, defaults to 2
        :type p: int, optional
        :param const: The constraints placed on the parameters A and b, defaults to None
        :type const: string, optional
        :param normalize: If true, the sample is normalized, defaults to True
        :type normalize: bool, optional
        :param d: The degree of polynomial features points are mapped to, defaults to 10
        :type d: int, optional
        :param kernelized: If True, the kernelized inverse christoffel function is constructed, defaults to False
        :type kernelized: bool, optional
        :param rho: Inverse Christoffel function constant rho, defaults to 0.0001
        :type rho: float, optional
        :param rbf: If True, a radical basis function kernel is used, defaults to False
        :type rbf: bool, optional
        :param scale: The length scale of the kernel, defaults to 1.0
        :type scale: float, optional
        :return: A :class:`dadra.Estimator` object
        :rtype: :class:`dadra.Estimator`
        :param iso_dim: Isolates the samples at the specified dimensions, currently only implemented when dyn_sys.all_time is True, defaults to None
        :type iso_dim: list, optional
        :return: A :class:`dadra.Estimator` object
        :rtype: :class:`dadra.Estimator`
        """
        sampler = Sampler(sample_fn, state_dim, timesteps, parts, all_time=all_time)
        return cls(
            sampler,
            epsilon=epsilon,
            delta=delta,
            christoffel=christoffel,
            p=p,
            const=const,
            normalize=normalize,
            d=d,
            kernelized=kernelized,
            rho=rho,
            rbf=rbf,
            scale=scale,
            iso_dim=iso_dim,
        )

    @classmethod
    def estimator_from_func(
        cls,
        dyn_func,
        intervals,
        state_dim,
        timesteps=100,
        parts=1001,
        epsilon=0.05,
        delta=1e-9,
        christoffel=False,
        p=2,
        const=None,
        normalize=True,
        d=10,
        kernelized=False,
        rho=0.0001,
        rbf=False,
        scale=0.5,
        iso_dim=None,
    ):
        """Class method that allows for an instance of :class:`dadra.Estimator` to be initialized
        using a dynamic function, and the components for an instance of :class:`SimpleSystem`
        rather than explicitly passing in a :class:`System` object

        :param dyn_func: The function defining the dynamics of the system
        :type dyn_func: function
        :param intervals: The intervals corresponding to the possible values of the initial states of the variables in the system
        :type intervals: list
        :param state_dim: The degrees of freedom of the system
        :type state_dim: int
        :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
        :type timesteps: int, optional
        :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
        :type parts: int, optional
        :param epsilon: The accuracy parameter, defaults to 0.05
        :type epsilon: float, optional
        :param delta: The confidence parameter, defaults to 1e-9
        :type delta: float, optional
        :param christoffel: If True, uses Christoffel functions to estimate the reachable set. If False, the p-norm method is used, defaults to False
        :type christoffel: bool, optional
        :param p: The order of p-norm, defaults to 2
        :type p: int, optional
        :param const: The constraints placed on the parameters A and b, defaults to None
        :type const: string, optional
        :param normalize: If true, the sample is normalized, defaults to True
        :type normalize: bool, optional
        :param d: The degree of polynomial features points are mapped to, defaults to 10
        :type d: int, optional
        :param kernelized: If True, the kernelized inverse christoffel function is constructed, defaults to False
        :type kernelized: bool, optional
        :param rho: Inverse Christoffel function constant rho, defaults to 0.0001
        :type rho: float, optional
        :param rbf: If True, a radical basis function kernel is used, defaults to False
        :type rbf: bool, optional
        :param scale: The length scale of the kernel, defaults to 1.0
        :type scale: float, optional
        :return: A :class:`dadra.Estimator` object
        :rtype: :class:`dadra.Estimator`
        :param iso_dim: Isolates the samples at the specified dimensions, currently only implemented when dyn_sys.all_time is True, defaults to None
        :type iso_dim: list, optional
        :return: A :class:`dadra.Estimator` object
        :rtype: :class:`dadra.Estimator`
        """
        dyn_sys = SimpleSystem(dyn_func, intervals, state_dim, timesteps, parts)
        return cls(
            dyn_sys,
            epsilon=epsilon,
            delta=delta,
            christoffel=christoffel,
            p=p,
            const=const,
            normalize=normalize,
            d=d,
            kernelized=kernelized,
            rho=rho,
            rbf=rbf,
            scale=scale,
            iso_dim=iso_dim,
        )

    @classmethod
    def estimator_from_list(
        cls,
        func_list,
        intervals,
        timesteps=100,
        parts=1001,
        epsilon=0.05,
        delta=1e-9,
        christoffel=False,
        p=2,
        const=None,
        normalize=True,
        d=10,
        kernelized=False,
        rho=0.0001,
        rbf=False,
        scale=0.5,
        iso_dim=None,
    ):
        """Class method that allows for an instance of :class:`dadra.Estimator` to be initialized
        using a list of dynamic functions, and the components for an instance of :class:`SimpleSystem`
        rather than explicitly passing in a :class:`System` object

        :param dyn_func: The list of functions, one for each variable, that define the dynamics of the system
        :type dyn_func: function
        :param intervals: The intervals corresponding to the possible values of the initial states of the variables in the system
        :type intervals: list
        :param timesteps: The number of timesteps over which to compute the sample, defaults to 100
        :type timesteps: int, optional
        :param parts: The number of parts to partition the time interval into for computing the sample, defaults to 1001
        :type parts: int, optional
        :param epsilon: The accuracy parameter, defaults to 0.05
        :type epsilon: float, optional
        :param delta: The confidence parameter, defaults to 1e-9
        :type delta: float, optional
        :param christoffel: If True, uses Christoffel functions to estimate the reachable set. If False, the p-norm method is used, defaults to False
        :type christoffel: bool, optional
        :param p: The order of p-norm, defaults to 2
        :type p: int, optional
        :param const: The constraints placed on the parameters A and b, defaults to None
        :type const: string, optional
        :param normalize: If true, the sample is normalized, defaults to True
        :type normalize: bool, optional
        :param d: The degree of polynomial features points are mapped to, defaults to 10
        :type d: int, optional
        :param kernelized: If True, the kernelized inverse christoffel function is constructed, defaults to False
        :type kernelized: bool, optional
        :param rho: Inverse Christoffel function constant rho, defaults to 0.0001
        :type rho: float, optional
        :param rbf: If True, a radical basis function kernel is used, defaults to False
        :type rbf: bool, optional
        :param scale: The length scale of the kernel, defaults to 1.0
        :type scale: float, optional
        :param iso_dim: Isolates the samples at the specified dimensions, currently only implemented when dyn_sys.all_time is True, defaults to None
        :type iso_dim: list, optional
        :return: A :class:`dadra.Estimator` object
        :rtype: :class:`dadra.Estimator`
        """
        dyn_sys = SimpleSystem.get_system(func_list, intervals, timesteps, parts)
        return cls(
            dyn_sys,
            epsilon=epsilon,
            delta=delta,
            christoffel=christoffel,
            p=p,
            const=const,
            normalize=normalize,
            d=d,
            kernelized=kernelized,
            rho=rho,
            rbf=rbf,
            scale=scale,
            iso_dim=iso_dim,
        )

    def check_sys(self):
        """Checks whether the dynamic system passed to the constructor is valid

        :raises TypeError: If an object of type other than :class:`dadra.System` is passed
        """
        if not isinstance(self.dyn_sys, System):
            raise TypeError("Object of type System must be passed")

    def get_num_samples(self, iso=False):
        """Compute the number of samples needed to satisfy the specified probabilistic guarantees

        :return: The number of samples needed to satisfy the specified probabilistic guarantees
        :rtype: int
        """
        if not iso:
            n_x = self.dyn_sys.state_dim
        else:
            n_x = len(self.iso_dim)
        if not self.christoffel:
            return p_num_samples(self.epsilon, self.delta, n_x, self.const)
        else:
            return c_num_samples(self.epsilon, self.delta, n_x, self.d)

    def get_sample(self):
        """Draws the number of samples necessary to satisfy the specified probabilistic guarantees

        :return: Array of samples
        :rtype: numpy.ndarray
        """
        print(f"Drawing {self.num_samples} samples")
        start_sampling = time.perf_counter()
        samples = self.dyn_sys.sample_system(self.num_samples)
        end_sampling = time.perf_counter()

        print(
            f"Time to draw {self.num_samples} samples: {format_time(end_sampling - start_sampling)}"
        )

        if self.normalize:
            samples = (samples - np.mean(samples, axis=0)) / np.std(samples)

        return samples

    def estimate(self):
        """Both draws the samples and makes an estimate of the reachable set"""
        self.sample_system()
        self.compute_estimate()

    def sample_system(self):
        """Draws and stores samples from the system"""
        self.samples = self.get_sample()
        print(f"Shape of all samples: {self.samples.shape}")

        if self.iso_dim is not None:
            self.iso_samples = self.samples[: self.num_iso_samples, :, self.iso_dim]
            print(f"Shape of dimension reduced samples: {self.iso_samples.shape}")

    def set_iso_dim(self, iso_dim):
        """Sets the iso_dim attribute and updates iso_samples attribute to correspond to the new iso_dim

        :param iso_dim: Isolates the samples at the specified dimensions
        :type iso_dim: list, optional
        """
        self.iso_dim = (
            iso_dim if isinstance(iso_dim, list) or iso_dim is None else [iso_dim]
        )
        if self.iso_dim is not None:
            self.num_iso_samples = self.get_num_samples(iso=True)
            print(f"Number of samples for isolated dimensions: {self.num_iso_samples}")
            self.iso_samples = self.samples[: self.num_iso_samples, :, self.iso_dim]

    def compute_estimate(self):
        """Computes an estimate of the reachable set based on the type of method specified"""
        if not self.christoffel:
            if not self.dyn_sys.all_time:
                self.A, self.b, self.status = self.solve_p()
            else:
                self.solution_list = self.solve_p()
        else:
            self.C, self.level = self.construct_christoffel()

    def solve_p(self):
        """Solves for the optimal p-norm ball that estimates the reachable set

        :return: The matrix corresponding to parameter A of the p-norm ball, the vector corresponding to parameter b of the p-norm ball, and the status of the optimization problem
        :rtype: tuple
        """
        print(f"Solving for optimal p-norm ball (p={self.p})")

        if self.iso_dim is None:
            curr_samples = self.samples
        else:
            print(
                f"Sovling for p-norm ball with respect to time at isolated dimensions: {self.iso_dim}"
            )
            curr_samples = self.iso_samples
            if len(curr_samples.shape) != 3:
                curr_samples = np.expand_dims(curr_samples, axis=2)

        start_time = time.perf_counter()
        if not self.dyn_sys.all_time:
            A, b, status = solve_p_norm(
                curr_samples, self.dyn_sys.state_dim, self.p, self.const
            )
            end_time = time.perf_counter()

            print(
                f"Time to solve for optimal p-norm ball: {format_time(end_time - start_time)}"
            )

            if status != "optimal":
                warnings.warn(
                    f"Solution for p-norm ball is not optimal: status = {status}"
                )

            return A, b, status

        else:
            solution_list = multi_p_norm(curr_samples, self.p, self.const)
            end_time = time.perf_counter()

            print(
                f"Time to solve for optimal p-norm ball for all timesteps: {format_time(end_time - start_time)}"
            )

            start_check = time.perf_counter()
            self.check_statuses(solution_list)
            end_check = time.perf_counter()

            print(
                f"Time to check each solution status: {format_time(end_check-start_check)}"
            )

            if self.num_opt != len(solution_list):
                warnings.warn(
                    "Not all p-norm ball solutions across the timesteps are optimal accurate"
                    + "\n"
                    + f"number of optimal solutions: {self.num_opt}"
                    + "\n"
                    + f"number of optimal inaccurate solutions: {self.num_opt_in}"
                    + "\n"
                    + f"number of infeasible or unbounded solutions: {self.num_fail}"
                )

            return solution_list

    def check_statuses(self, solution_list=None):
        """Checks the status of each of the solutions in the list of p-norm ball solutions

        :param solution_list: List of solutions where each solution is a tuple of the form (A, b, status), defaults to None
        :type solution_list: list, optional
        """
        self.num_opt = 0
        self.num_opt_in = 0
        self.num_fail = 0
        s_list = solution_list if solution_list is not None else self.solution_list
        for d in s_list:
            status = d["status"]
            if status == "optimal":
                self.num_opt += 1
            elif status == "optimal_inaccurate":
                self.num_opt_in += 1
            else:
                self.num_fail += 1

    def construct_christoffel(self):
        """Constructs the inverse Christoffel function and the respective level parameter from the samples

        :return: the inverse Christoffel function and the level parameter
        :rtype: tuple
        """
        if self.kernelized:
            C = construct_kernelized_inv_christoffel(
                self.samples, self.d, rho=self.rho, rbf=self.rbf, scale=self.scale
            )
        else:
            C = construct_inv_christoffel(self.samples, self.d)

        start_lvl_param = time.perf_counter()
        values = C(self.samples)
        level = np.max(values)
        end_lvl_param = time.perf_counter()
        print(
            f"Time to compute level parameter: {format_time(end_lvl_param - start_lvl_param)}"
        )
        return C, level

    def empirical_estimate(self, num_samples_emp=None):
        """Computes the ratio of samples within the estimated reachable set

        :param num_samples_emp: The number of samples to draw for the empirical estimate, defaults to the number of samples drawn to obtain the p-norm ball
        :type num_samples_emp: int, optional
        """
        if not self.christoffel and self.status != "optimal":
            warnings.warn(
                "Non-optimal solution to p-norm ball. Empirical estimate may be innaccurate"
            )

        if num_samples_emp is None:
            num_samples_emp = self.num_samples

        print(f"Drawing {num_samples_emp} samples for empirical estimate")
        start_sampling = time.perf_counter()
        samples_emp = self.dyn_sys.sample_system(num_samples_emp)
        end_sampling = time.perf_counter()

        print(
            f"Time to draw {num_samples_emp} samples: {format_time(end_sampling - start_sampling)}"
        )

        if self.normalize:
            samples_emp = (samples_emp - np.mean(samples_emp, axis=0)) / np.std(
                samples_emp
            )

        if not self.christoffel:
            ratio = p_emp_estimate(
                samples_emp, self.A, self.b, self.dyn_sys.state_dim, self.p
            )
        else:
            ratio = c_emp_estimate(self.samples, self.C, self.level)

        print(f"Ratio of samples within the estimated reachable set: {ratio}")

    def summary(self):
        """Prints a summary of the attributes of this instance of :class:`Estimator`"""
        summary_str = (
            "--------------------------------------------------------------------"
            + "\n"
        )
        summary_str += "Estimator Summary" + "\n"
        summary_str += (
            "===================================================================="
            + "\n"
        )
        summary_str += f"State dimension: {self.dyn_sys.state_dim}" + "\n"
        summary_str += f"Accuracy parameter epsilon: {self.epsilon}" + "\n"
        summary_str += f"Confidence parameter delta: {self.delta}" + "\n"
        summary_str += f"Number of samples: {self.num_samples}" + "\n"
        if not self.christoffel:
            summary_str += f"Method of estimation: p-Norm Ball" + "\n"
            summary_str += f"p-norm p value: {self.p}" + "\n"
            summary_str += f"Constraints on p-norm ball: {self.const}" + "\n"
            if not self.dyn_sys.all_time:
                if all([self.A is None, self.b is None, self.status is None]):
                    summary_str += (
                        f"Status of p-norm ball solution: No estimate has been made yet"
                        + "\n"
                    )
                else:

                    summary_str += (
                        f"Status of p-norm ball solution: {self.status}" + "\n"
                    )
            else:
                if self.iso_dim is not None:
                    summary_str += f"Isolated dimensions: {self.iso_dim}" + "\n"
                    summary_str += (
                        f"Number of samples for isolated dimensions: {self.num_iso_samples}"
                        + "\n"
                    )
                if self.solution_list is None:
                    summary_str += (
                        f"Status of p-norm ball solutions: No estimates have been made yet"
                        + "\n"
                    )
                else:
                    summary_str += (
                        f"Ratio of failed p-norm ball solutions: {self.num_fail / len(self.solution_list)}"
                        + "\n"
                    )
        else:
            summary_str += f"Method of estimation: Inverse Christoffel Function" + "\n"
            summary_str += f"Degree of polynomial features: {self.d}" + "\n"
            summary_str += f"Kernelized: {self.kernelized}" + "\n"
            summary_str += f"Constant rho: {self.rho}" + "\n"
            summary_str += f"Radical basis function kernel: {self.rbf}" + "\n"
            summary_str += f"Length scale of kernel: {self.scale}" + "\n"
            if all([self.C is None, self.level is None]):
                summary_str += (
                    f"Status of Christoffel Function estimate: No estimate has been made yet"
                    + "\n"
                )
            else:
                summary_str += f"Christoffel function level: {self.level}" + "\n"
                summary_str += (
                    f"Status of Christoffel function estimate: Estimate made" + "\n"
                )

        summary_str += (
            "--------------------------------------------------------------------"
            + "\n"
        )
        print(summary_str)

    def plot_samples(
        self,
        fig_name,
        gif_name=None,
        max_num_samples=100,
        figsize=(10, 10),
        color="default",
        **kwargs,
    ):
        """Plots the samples of 2-dimensional shape in 2D or samples of 3-dimensional shape in 3-dimensions drawn with respect to time

        :param fig_name: The name of the file to save the plot to
        :type fig_name: str
        :param gif_name: The name of the animated gif to be saved, only applicable for 3-dimensional plots when all_time is True, defaults to None
        :type gif_name: str, optional
        :param max_num_samples: The maximum number of samples to plot, defaults to 100
        :type max_num_samples: int, optional
        :param figsize: the size of the figure to be saved, defaults to (10, 10)
        :type figsize: tuple, optional
        :param color: The color to draw the lines in, defaults to "default"
        :type color: str, optional
        :raises ValueError: If the samples are of an incompatible shape
        """
        if not self.dyn_sys.all_time and self.iso_dim is None:
            plot_sample(self.samples[:max_num_samples], fig_name)
        else:
            if len(self.iso_samples.shape) == 2 or self.iso_samples.shape[2] == 1:
                if len(self.iso_samples.shape) != 2:
                    sample_squeezed = np.squeeze(self.iso_samples, axis=2)
                else:
                    sample_squeezed = self.iso_samples
                time_x = np.linspace(0, self.dyn_sys.timesteps, self.dyn_sys.parts)

                plot_sample_time(
                    time_x,
                    sample_squeezed[:max_num_samples],
                    fig_name,
                    figsize=figsize,
                    color=color,
                    **kwargs,
                )

            elif len(self.iso_samples.shape) == 3 and self.iso_samples.shape[2] == 3:
                grow_plot_3d(
                    self.iso_samples[:max_num_samples],
                    fig_name,
                    gif_name,
                    figsize,
                    **kwargs,
                )

            else:
                raise ValueError(
                    f"Cannot handle samples of shape: {self.iso_samples.shape}"
                )

    def plot_reachable_time(
        self,
        fig_name,
        gif_name=None,
        grid_n=200,
        num_samples_show=50,
        figsize=(10, 10),
        **kwargs,
    ):
        """Plots the reachable set estimate (as well as possibly some samples) over time. Currently only implmented for p-norm balls

        :param fig_name: The name of the file to save the plot to
        :type fig_name: str
        :param gif_name: The name of the animated gif to be saved, only applicable for 3-dimensional plots, defaults to None
        :type gif_name: str, optional
        :param grid_n: The number of points to test for the p-norm ball estimation at each time step, defaults to 200
        :type grid_n: int, optional
        :param num_samples_show: The number of sample trajectories to show in addition to the reachable set, defaults to 50
        :type num_samples_show: int, optional
        :param figsize: the size of the figure to be saved, defaults to (10, 10)
        :type figsize: tuple, optional
        """
        if not self.dyn_sys.all_time:
            warnings.warn(
                "`plot_reachable_time` is meant for samples across all timesteps"
            )

        if self.iso_dim is None:
            curr_samples = self.samples
        else:
            curr_samples = self.iso_samples

        if len(curr_samples.shape) == 2 or curr_samples.shape[2] == 1:
            if len(curr_samples.shape) != 2:
                curr_samples = np.squeeze(curr_samples, axis=2)

        if len(curr_samples.shape) == 2:
            min_vals, max_vals = [], []
            for i in range(self.dyn_sys.parts):
                A_i, b_i = self.solution_list[i]["A"], self.solution_list[i]["b"]
                vals_i = p_compute_vals(curr_samples[:, i], A_i, b_i, grid_n=grid_n)
                min_vals.append(min(vals_i))
                max_vals.append(max(vals_i))

            time_x = np.linspace(0, self.dyn_sys.timesteps, self.dyn_sys.parts)

            plot_reach_time(
                time_x,
                curr_samples,
                min_vals,
                max_vals,
                fig_name,
                num_samples_show=num_samples_show,
                figsize=figsize,
                **kwargs,
            )

        elif len(curr_samples.shape) == 3 and curr_samples.shape[2] == 3:
            dict_list = p_dict_list(
                curr_samples,
                self.solution_list,
                self.dyn_sys.parts,
                num_indices=20,
                logspace=True,
                grid_n=grid_n,
            )
            plot_contour_3D_time(
                curr_samples[:num_samples_show],
                dict_list,
                fig_name,
                gif_name,
                figsize,
                **kwargs,
            )

    def plot_2D_cont(self, fig_name, grid_n=200):
        """Computes the contours of the reachable set and plots them in 2D

        :param fig_name: The name of the file to save the plot to
        :type fig_name: str
        :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 200
        :type grid_n: int
        """
        if not self.christoffel:
            cont_compute = partial(
                p_compute_contour_2D,
                sample=self.samples,
                A_val=self.A,
                b_val=self.b,
                n_x=self.dyn_sys.state_dim,
                p=self.p,
                grid_n=grid_n,
            )

            print("Computing 2D contours")
            cont_start = time.perf_counter()
            xv1, yv1, z_cont, z_min, z_max = cont_compute(cont_axis=2)
            xv2, zv1, y_cont, y_min, y_max = cont_compute(cont_axis=1)
            yv2, zv2, x_cont, x_min, x_max = cont_compute(cont_axis=0)
            cont_end = time.perf_counter()
            print(f"Time to compute 2D contours: {format_time(cont_end - cont_start)}")

            plot_contour_2D(
                xv1,
                yv1,
                z_cont,
                z_max,
                self.samples,
                fig_name,
                xv2=xv2,
                zv1=zv1,
                y_cont=y_cont,
                y_level=y_max,
                yv2=yv2,
                zv2=zv2,
                x_cont=x_cont,
                x_level=x_max,
            )

        else:
            start_comp_cont = time.perf_counter()
            xv, yv, cont = c_compute_contours(self.C, self.samples, grid_n)
            end_comp_cont = time.perf_counter()
            print(
                f"Time to compute contour: {format_time(end_comp_cont - start_comp_cont)}"
            )

            plot_contour_2D(xv, yv, cont, self.level, self.samples, fig_name)

    def plot_3D_cont(self, fig_name, grid_n=100, gif_name=None):
        """Computes and plots the contours in 3D with the option for saving an animated gif of the rotating graph

        :param fig_name: The name of the file to save the plot to
        :type fig_name: str
        :param grid_n: The side length of the cube of points to be used for computing contours, defaults to 100
        :type grid_n: int
        :param gif_name: The name of the file to save the gif to, defaults to None
        :type gif_name: str, optional
        """
        print("Computing 3D contour")
        cont_start = time.perf_counter()
        d0, d1, cont_min, cont_max, c_min, c_max = p_compute_contour_3D(
            self.samples, self.A, self.b, grid_n=grid_n
        )
        cont_end = time.perf_counter()
        print(f"Time to compute 3D contour: {format_time(cont_end - cont_start)}")

        plot_contour_3D(
            d0,
            d1,
            cont_min,
            c_min,
            c_max,
            self.samples,
            fig_name,
            gif_name=gif_name,
            z_cont2=cont_max,
        )



import numpy as np
import time


def format_time(t):
    """Formats time into string with minutes and seconds

    :param t: The time in seconds
    :type t: float
    :return: Formatted string with minutes and seconds
    :rtype: string
    """
    time_string = time.strftime("%M:%S", time.gmtime(t))
    time_list = time_string.split(":")
    num_min = time_list[0]
    num_sec = time_list[1]
    return f"{num_min} minutes and {num_sec} seconds"


def normalize0(X):
    """Normalizes the input array with respect to the square root of the first axis

    :param X: Input array to be normalized
    :type X: numpy.ndarray
    :return: The normalized array
    :rtype: numpy.ndarray
    """
    return X / np.sqrt(X.shape[0])


def normalize1(X):
    """Normalizes the input array with respect to the average absolute values

    :param X: Input array to be normalized
    :type X: numpy.ndarray
    :return: The normalized array
    :rtype: numpy.ndarray
    """
    return X / np.abs(X).mean()


def normalize2(X):
    """Normalizes the input array with respect to the maximum of the absolute values

    :param X: Input array to be normalized
    :type X: numpy.ndarray
    :return: The normalized array
    :rtype: numpy.ndarray
    """
    return X / np.max(np.abs(X))


def normalize(X, mu=0, std=1):
    """Normalizes the input array with mean mu and standard deviation std

    :param X: Input array to be normalized
    :type X: numpy.ndarray
    :param mu: The mean, defaults to 0
    :type mu: int, optional
    :param std: The standard deviation, defaults to 1
    :type std: int, optional
    :return: The normalized array
    :rtype: nnumpy.ndarray
    """
    return (X - mu) / std



import numpy as np

from dadra.disturbance import Disturbance


def duffing_oscillator(
    y, t, disturbance: Disturbance = None, eta=1.0, alpha=0.05, omega=1.3, gamma=0.4
):
    """Defines the dynamics of a Duffing oscillator based on initial states

    :param y: The initial states of the Duffing oscillator (2 dimensions)
    :type y: numpy.ndarray
    :param t: The time at which the derivatives of the system are computed
    :type t: int
    :param disturbance: The disturbance to be added to each dimension, defaults to None
    :type disturbance: :class:`dadra.Disturbance`, optional
    :param eta: The weight of the disturbance to be added to each dimension, defaults to 1.0
    :type eta: float, optional
    :param alpha: Duffing oscillator constant alpha, defaults to 0.05
    :type alpha: float, optional
    :param omega: Duffing oscillator constant omega, defaults to 1.3
    :type omega: float, optional
    :param gamma: Duffing oscillator constant gamma, defaults to 0.4
    :type gamma: float, optional
    :return: The partial derivatives of the Duffing oscillator along x and y
    :rtype: list
    """
    if disturbance is None:
        dydt = [y[1], -alpha * y[1] + y[0] - y[0] ** 3 + gamma * np.cos(omega * t)]
    else:
        dydt = [
            y[1] + disturbance.get_dist(0, t) * eta,
            -alpha * y[1]
            + y[0]
            - y[0] ** 3
            + gamma * np.cos(omega * t)
            + disturbance.get_dist(1, t) * eta,
        ]

    return dydt


def lorenz_system(
    z, t, disturbance: Disturbance = None, eta=1.0, sigma=10.0, rho=28.0, beta=8 / 3
):
    """Defines the dynamics of a Lorenz system based on initial states

    :param z: The initial states of the Lorenz system (3 dimensions)
    :type z: numpy.ndarray
    :param t: The time at which the derivatives of the system are computed
    :type t: int
    :param disturbance: The disturbance to be added to each dimension, defaults to None
    :type disturbance: :class:`dadra.Disturbance`, optional
    :param eta: The weight of the disturbance to be added to each dimension, defaults to 1.0
    :type eta: float, optional
    :param sigma: Lorenz system constant sigma, defaults to 10.
    :type sigma: float, optional
    :param rho: Lorenz system constant rho, defaults to 28.
    :type rho: float, optional
    :param beta: Lorenz system constant beta, defaults to 8/3
    :type beta: float, optional
    :return: The partial derivatives of the Lorenz system along x, y, and z
    :rtype: list
    """
    if disturbance is None:
        dzdt = [
            sigma * (z[1] - z[0]),
            z[0] * (rho - z[2]) - z[1],
            z[0] * z[1] - beta * z[2],
        ]
    else:
        dzdt = [
            sigma * (z[1] - z[0]) + disturbance.get_dist(0, t) * eta,
            z[0] * (rho - z[2]) - z[1] + disturbance.get_dist(1, t) * eta,
            z[0] * z[1] - beta * z[2] + disturbance.get_dist(2, t) * eta,
        ]
    return dzdt


def planar_quadrotor(y, t, u1, u2, g=9.81, K=0.89 / 1.4, d0=70, d1=17, n0=55):
    """Defines the dynamics of a Planar Quadrotor Model based on initial states

    :param y: The initial states of the model (6 states: x, h, theta, and their first derivatives)
    :type y: numpy.ndarray
    :param t: The time at which the derivatives of the system are computed
    :type t: int
    :param u1: System input 1, treated as disturbance, the motor thrust
    :type u1: float
    :param u2: System input 2, treated as disturbance, the desired angle
    :type u2: float
    :param g: System constant g, defaults to 9.81
    :type g: float, optional
    :param K: System constant K, defaults to 0.89/1.4
    :type K: float, optional
    :param d0: System constant d0, defaults to 70
    :type d0: int, optional
    :param d1: System constant d1, defaults to 17
    :type d1: int, optional
    :param n0: System constant n0, defaults to 55
    :type n0: int, optional
    :return: The partial derivatives of the Planar Quadrotor Model system along each of the 6 states.
    :rtype: list
    """
    x, xdot, h, hdot, theta, thetadot = y
    dydt = [
        xdot,
        u1 * K * np.sin(theta),
        hdot,
        -g + u1 * K * np.cos(theta),
        thetadot,
        -d0 * theta - d1 * thetadot + n0 * u2,
    ]
    return dydt


def traffic(y, t, d=0):
    """Defines the dynamics of a Monotone Traffic Model based on initial states

    :param y: The initial states of the model (n-dimensional)
    :type y: numpy.ndarray
    :param t: The time at which the derivatives of the system are computed
    :type t: int
    :param d: disturbances, defaults to 0
    :type d: int, optional
    :return: The partial derivatives of the Monotone Traffic Model along each of the n dimensions
    :rtype: numpy.ndarray
    """
    v = 0.5  # free-flow speed, in links/period
    w = 1 / 6  # congestion-wave speed, in links/period
    c = 40  # capacity (max downstream flow), in vehicles/period
    xbar = 320  # max occupancy when jammed, in vehicles
    b = 1  # fraction of vehicle staying on the network after each link
    T = 30  # time step for the continuous-time model
    nx = np.size(y)
    dy = np.zeros(np.shape(y))
    dy[0] = 1 / T * (d - min(c, v * y[0], 2 * w * (xbar - y[1])))
    for i in range(1, nx - 1):
        dy[i] = (
            1
            / T
            * (
                b * min(c, v * y[i - 1], w * (xbar - y[i]))
                - min(c, v * y[i], w / b * (xbar - y[i + 1]))
            )
        )
    dy[-1] = 1 / T * (b * min(c, v * y[-2], w * (xbar - y[-1])) - min(c, v * y[-1]))
    return dy


def quadrotor(x, t, g=9.81, R=0.1, l=0.5, M_rotor=0.1, M=1.0, u1=1.0, u2=0.0, u3=0.0):
    """Defines the dynamics of a 12-state Quadrotor Model based on initial states

    :param x: 12 state input vector
    :type x: numpy.array
    :param t: The time
    :type t: [type]
    :param g: The gravity constant, defaults to 9.81
    :type g: float, optional
    :param R: The radius of center mass, defaults to 0.1
    :type R: float, optional
    :param l: The distance of motors to center mass, defaults to 0.5
    :type l: float, optional
    :param M_rotor: The motor mass, defaults to 0.1
    :type M_rotor: float, optional
    :param M: The center mass, defaults to 1.0
    :type M: float, optional
    :param u1: The desired height, defaults to 1.0
    :type u1: float, optional
    :param u2: The desired roll, defaults to 0.0
    :type u2: float, optional
    :param u3: The desired pitch, defaults to 0.0
    :type u3: float, optional
    :return: The partial derivatives of the 12-state quadrotor model
    :rtype: list
    """
    x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12 = x

    # total mass
    m = M + 4.0 * M_rotor

    # moments of inertia
    Jx = 2.0 / 5.0 * M * (R ** 2) + 2.0 * (l ** 2) * M_rotor
    Jy = Jx
    Jz = 2.0 / 5.0 * M * (R ** 2) + 4.0 * (l ** 2) * M_rotor

    # height control
    F = (m * g - 10.0 * (x3 - u1)) + 3.0 * x6

    # roll control
    tau_phi = -(x7 - u2) - x10

    # pitch control
    tau_theta = -(x8 - u3) - x11

    # heading uncontrolled
    tau_psi = 0

    dx1 = (
        np.cos(x8) * np.cos(x9) * x4
        + (np.sin(x7) * np.sin(x8) * np.cos(x9) - np.cos(x7) * np.sin(x9)) * x5
        + (np.cos(x7) * np.sin(x8) * np.cos(x9) + np.sin(x7) * np.sin(x9)) * x6
    )
    dx2 = (
        np.cos(x8) * np.sin(x9) * x4
        + (np.sin(x7) * np.sin(x8) * np.sin(x9) + np.cos(x7) * np.cos(x9)) * x5
        + (np.cos(x7) * np.sin(x8) * np.sin(x9) - np.sin(x7) * np.cos(x9)) * x6
    )
    dx3 = np.sin(x8) * x4 - np.sin(x7) * np.cos(x8) * x5 - np.cos(x7) * np.cos(x8) * x6
    dx4 = x12 * x5 - x11 * x6 - g * np.sin(x8)
    dx5 = x10 * x6 - x12 * x4 + g * np.cos(x8) * np.sin(x7)
    dx6 = x11 * x4 - x10 * x5 + g * np.cos(x8) * np.cos(x7) - F / m
    dx7 = x10 + np.sin(x7) * np.tan(x8) * x11 + np.cos(x7) * np.tan(x8) * x12
    dx8 = np.cos(x7) * x11 - np.sin(x7) * x12
    dx9 = np.sin(x7) / np.cos(x8) * x11 + np.cos(x7) / np.cos(x8) * x12
    dx10 = (Jy - Jz) / Jx * x11 * x12 + 1 / Jx * tau_phi
    dx11 = (Jz - Jx) / Jy * x10 * x12 + 1 / Jy * tau_theta
    dx12 = (Jx - Jy) / Jz * x10 * x11 + 1 / Jz * tau_psi

    dx = [dx1, dx2, dx3, dx4, dx5, dx6, dx7, dx8, dx9, dx10, dx11, dx12]

    return dx
